from importlib import import_module
from pathlib import Path

versions = {}
for subpath in Path(__file__).parent.glob("*.py"):
    name = subpath.stem
    if name == __name__:
        continue
    submod = import_module(__package__ + "." + name)
    if hasattr(submod, "version"):
        versions[submod.version] = submod

max_version = max(versions.keys())



from uuid import uuid4

from cached_property import cached_property

from ops.framework import (
    StoredState,
    EventBase,
    EventSource,
    ObjectEvents,
)
from ops.model import ModelError

from .base import VersionedInterface


class LBProviderAvailable(EventBase):
    pass


class LBResponseAvailable(EventBase):
    pass


class LBResponseChanged(EventBase):
    pass


class LBProviderEvents(ObjectEvents):
    available = EventSource(LBProviderAvailable)
    response_available = EventSource(LBResponseAvailable)
    response_changed = EventSource(LBResponseChanged)


class LBProvider(VersionedInterface):
    """API used to interact with the provider of loadbalancers."""

    state = StoredState()
    on = LBProviderEvents()

    def __init__(self, charm, relation_name):
        super().__init__(charm, relation_name)
        self.relation_name = relation_name
        # just call this to enforce that only one app can be related
        self.model.get_relation(relation_name)
        self.state.set_default(
            response_hashes={}, was_available=False, was_response_available=False
        )

        for event in (
            charm.on[relation_name].relation_created,
            charm.on[relation_name].relation_joined,
            charm.on[relation_name].relation_changed,
            charm.on[relation_name].relation_departed,
            charm.on[relation_name].relation_broken,
        ):
            self.framework.observe(event, self._check_provider)

    def _check_provider(self, event):
        if self.is_available:
            if not self.state.was_available:
                self.state.was_available = True
                self.on.available.emit()
            if not self.state.was_response_available:
                self.state.was_response_available = True
                self.on.response_available.emit()
            if self.is_changed:
                self.on.response_changed.emit()
        elif self.state.was_available:
            self.state.was_available = False
            self.state.was_response_available = False
            if self.state.response_hashes:
                self.state.response_hashes = {}
                self.on.response_changed.emit()

    @property
    def relation(self):
        return self.relations[0] if self.relations else None

    def get_request(self, name):
        """Get or create a Load Balancer Request object.

        May raise a ModelError if unable to create a request.
        """
        if not self.charm.unit.is_leader():
            raise ModelError("Unit is not leader")
        if not self.relation:
            raise ModelError("Relation not available")
        schema = self._schema(self.relation)
        local_data = self.relation.data[self.app]
        remote_data = self.relation.data[self.relation.app]
        request_key = "request_" + name
        response_key = "response_" + name
        if request_key in local_data:
            request_sdata = local_data[request_key]
            response_sdata = remote_data.get(response_key)
            request = schema.Request.loads(request_sdata, response_sdata)
        else:
            request = schema.Request()
            request.name = name
            request.id = uuid4().hex
        return request

    def get_response(self, name):
        """Get a specific Load Balancer Response by name.

        This is equivalent to `get_request(name).response`, except that it
        will return `None` if the response is not available.
        """
        if not self.is_available:
            return None
        request = self.get_request(name)
        if not request.response:
            return None
        return request.response

    def send_request(self, request):
        """Send a specific request.

        May raise a ModelError if unable to send the request.
        """
        if not self.charm.unit.is_leader():
            raise ModelError("Unit is not leader")
        if not self.relation:
            raise ModelError("Relation not available")
        # The sent_hash is used to tell when the provider's response has been
        # updated to match our request. We can't used the request hash computed
        # on the providing side because it may not match due to default values
        # being filled in on that side (e.g., the backend addresses).
        request.sent_hash = request.hash
        key = "request_" + request.name
        self.relation.data[self.app][key] = request.dumps()

    def remove_request(self, name):
        """Remove a specific request.

        May raise a ModelError if unable to remove the request.
        """
        if not self.charm.unit.is_leader():
            raise ModelError("Unit is not leader")
        if not self.relation:
            return
        key = "request_" + name
        self.relation.data[self.app].pop(key, None)
        self.state.response_hashes.pop(name, None)

    @property
    def all_requests(self):
        """A list of all requests which have been made."""
        requests = []
        if self.relation:
            for key in sorted(self.relation.data[self.app].keys()):
                if not key.startswith("request_"):
                    continue
                requests.append(self.get_request(key[len("request_") :]))
        return requests

    @property
    def revoked_responses(self):
        """A list of responses which are no longer available."""
        return [
            request.response
            for request in self.all_requests
            if not request.response and request.name in self.state.response_hashes
        ]

    @cached_property
    def all_responses(self):
        """A list of all responses which are available."""
        return [request.response for request in self.all_requests if request.response]

    @cached_property
    def complete_responses(self):
        """A list of all responses which are up to date with their associated
        request.
        """
        return [
            request.response
            for request in self.all_requests
            if request.response.received_hash == request.sent_hash
        ]

    @property
    def new_responses(self):
        """A list of complete responses which have not yet been acknowledged as
        handled or which have changed.
        """
        acked_responses = self.state.response_hashes
        return [
            response
            for response in self.complete_responses
            if response.hash != acked_responses.get(response.name)
        ]

    def ack_response(self, response):
        """Acknowledge that a given response has been handled.

        Can be called on a revoked response as well to remove it
        from the revoked_responses list.
        """
        if response:
            self.state.response_hashes[response.name] = response.hash
        else:
            self.state.response_hashes.pop(response.name, None)
        if not self.is_changed:
            try:
                from charms.reactive import clear_flag

                prefix = "endpoint." + self.relation_name
                clear_flag(prefix + ".response.changed")
            except ImportError:
                pass  # not being used in a reactive charm

    @property
    def is_changed(self):
        return self.new_responses or self.revoked_responses

    @property
    def is_available(self):
        return bool(self.relation)

    @property
    def has_response(self):
        return bool(self.complete_responses)

    @property
    def can_request(self):
        return self.is_available and self.unit.is_leader()

    def manage_flags(self):
        """Used to interact with charms.reactive-base charms."""
        from charms.reactive import toggle_flag

        prefix = "endpoint." + self.relation_name
        toggle_flag(prefix + ".available", self.is_available)
        toggle_flag(prefix + ".response.available", self.has_response)
        toggle_flag(prefix + ".response.changed", self.is_changed)



import weakref
from contextlib import contextmanager
from pathlib import Path

import pytest

from pytest_operator import OperatorTest


try:
    from importlib.resources import path as resource_path
except ImportError:
    # Shims for importlib_resources and pkg_resources to behave like the stdlib
    # version from 3.7+.
    try:
        from importlib_resources import path as resource_path
    except ImportError:
        from pkg_resources import resource_filename

        @contextmanager
        def resource_path(package, resource):
            rf = resource_filename(package, resource)
            yield Path(rf)


def pytest_configure(config):
    config.addinivalue_line("markers", "lb_charms")


@pytest.fixture(autouse=True, scope="class")
def lb_charms(request):
    """Fixture which injects the lb_charms attribute to marked OperatorTests.

    Any subclass of OperatorTest which is marked with @pytest.mark.lb_charms
    will have a lb_charms attribute injected into it which has the following
    attributes:

      * lb_provider - A charm which provides loadbalancers
      * lb_consumer - A charm which requires loadbalancers
      * lb_provider_reactive - A reactive charm which provides loadbalancers
      * lb_consumer_reactive - A reactive charm which requires loadbalancers
    """
    marker = request.node.get_closest_marker("lb_charms")
    if not marker:
        return
    if not issubclass(getattr(request, "cls", None), OperatorTest):
        return
    request.cls.lb_lib_url = "loadbalancer-interface"
    request.cls.lb_charms = property(lambda test_case: LBCharms(test_case))


class LBCharms:
    def __init__(self, test_case):
        self._test_case = weakref.proxy(test_case)
        self._lb_provider = None
        self._lb_consumer = None
        self._lb_provider_reactive = None
        self._lb_consumer_reactive = None

    def _render(self, charm_resource):
        with resource_path("loadbalancer_interface", "examples") as rp:
            return self._test_case.render_charm(
                rp / charm_resource,
                include=["wheelhouse.txt", "requirements.txt"],
                lb_lib_url=self._test_case.lb_lib_url,
            )

    @property
    def lb_provider(self):
        if self._lb_provider is None:
            self._lb_provider = self._render("lb-provider")
        return self._lb_provider

    @property
    def lb_consumer(self):
        if self._lb_consumer is None:
            self._lb_consumer = self._render("lb-consumer")
        return self._lb_consumer

    @property
    def lb_provider_reactive(self):
        if self._lb_provider_reactive is None:
            self._lb_provider_reactive = self._render("lb-provider-reactive")
        return self._lb_provider_reactive

    @property
    def lb_consumer_reactive(self):
        if self._lb_consumer_reactive is None:
            self._lb_consumer_reactive = self._render("lb-consumer-reactive")
        return self._lb_consumer_reactive



from operator import attrgetter

from cached_property import cached_property

from ops.framework import (
    StoredState,
    EventBase,
    EventSource,
    ObjectEvents,
)

from .base import VersionedInterface


class LBRequestsChanged(EventBase):
    pass


class LBConsumersEvents(ObjectEvents):
    requests_changed = EventSource(LBRequestsChanged)


class LBConsumers(VersionedInterface):
    """API used to interact with consumers of a loadbalancer provider."""

    state = StoredState()
    on = LBConsumersEvents()

    def __init__(self, charm, relation_name):
        super().__init__(charm, relation_name)
        self.relation_name = relation_name
        self.state.set_default(known_requests={})

        for event in (
            charm.on[relation_name].relation_created,
            charm.on[relation_name].relation_joined,
            charm.on[relation_name].relation_changed,
        ):
            self.framework.observe(event, self._check_consumers)

    def _check_consumers(self, event):
        if self.is_changed:
            self.on.requests_changed.emit()

    @cached_property
    def all_requests(self):
        """A list of all current consumer requests."""
        if not self.unit.is_leader():
            # Only the leader can process requests, so avoid mistakes
            # by not even reading the requests if not the leader.
            return []
        requests = []
        for relation in self.relations:
            schema = self._schema(relation)
            local_data = relation.data[self.app]
            remote_data = relation.data[relation.app]
            for key, request_sdata in sorted(remote_data.items()):
                if not key.startswith("request_"):
                    continue
                name = key[len("request_") :]
                response_sdata = local_data.get("response_" + name)
                request = schema.Request.loads(request_sdata, response_sdata)
                request.relation = relation
                if not request.backends:
                    for unit in sorted(relation.units, key=attrgetter("name")):
                        addr = relation.data[unit].get("ingress-address")
                        if addr:
                            request.backends.append(addr)
                requests.append(request)
                self.state.known_requests.setdefault(request.id, None)
        return requests

    @property
    def new_requests(self):
        """A list of requests with changes or no response."""
        return [
            request
            for request in self.all_requests
            if request.hash != self.state.known_requests[request.id]
        ]

    @property
    def removed_requests(self):
        """A list of requests which have been removed, either explicitly or
        because the relation was removed.
        """
        current_ids = {request.id for request in self.all_requests}
        unknown_ids = self.state.known_requests.keys() - current_ids
        schema = self._schema()
        removed_requests = []
        for req_id in sorted(unknown_ids):
            request = schema.Request()
            request.id = req_id
            removed_requests.append(request)
        return removed_requests

    def send_response(self, request):
        """Send a specific request's response."""
        request.response.received_hash = request.sent_hash
        key = "response_" + request.name
        request.relation.data[self.app][key] = request.response.dumps()
        self.state.known_requests[request.id] = request.hash
        if not self.new_requests:
            try:
                from charms.reactive import clear_flag

                prefix = "endpoint." + self.relation_name
                clear_flag(prefix + ".requests_changed")
            except ImportError:
                pass  # not being used in a reactive charm

    def revoke_response(self, request):
        """Revoke / remove the response for a given request."""
        if request.id:
            self.state.known_requests.pop(request.id, None)
        if request.relation:
            key = "response_" + request.name
            request.relation.data.get(self.app, {}).pop(key, None)

    @property
    def is_changed(self):
        return bool(self.new_requests or self.removed_requests)

    def manage_flags(self):
        """Used to interact with charms.reactive-base charms."""
        from charms.reactive import toggle_flag

        prefix = "endpoint." + self.relation_name
        toggle_flag(prefix + ".requests_changed", self.is_changed)



#!/usr/bin/env python3

import logging

from ops.charm import CharmBase
from ops.main import main
from ops.model import ActiveStatus, BlockedStatus, MaintenanceStatus, WaitingStatus

from loadbalancer_interface import LBProvider


log = logging.getLogger(__name__)


class RequiresOperatorCharm(CharmBase):
    def __init__(self, *args):
        super().__init__(*args)
        self.lb_provider = LBProvider(self, "lb-provider")

        if not self.lb_provider.is_available:
            self.unit.status = WaitingStatus("waiting on provider")

        self.framework.observe(self.lb_provider.on.available, self._request_lb)
        self.framework.observe(self.lb_provider.on.response_changed, self._get_lb)
        if self.lb_provider.is_available:
            self.framework.observe(self.on.config_changed, self._request_lb)

    def _request_lb(self, event):
        self.unit.status = MaintenanceStatus("sending request")
        request = self.lb_provider.get_request("my-service")
        request.protocol = request.protocols.https
        request.port_mapping = {443: 443}
        request.public = self.config["public"]
        self.lb_provider.send_request(request)
        self.unit.status = WaitingStatus("waiting on provider response")

    def _get_lb(self, event):
        response = self.lb_provider.get_response("my-service")
        if not response:
            return
        if response.error:
            self.unit.status = BlockedStatus(f"LB failed: {response.error}")
            log.error(
                f"LB failed ({response.error}):\n"
                f"{response.error_message}\n"
                f"{response.error_fields}"
            )
            return
        log.info(f"LB is available at {response.address}")
        self.lb_provider.ack_response(response)
        self.unit.status = ActiveStatus(response.address)


if __name__ == "__main__":
    main(RequiresOperatorCharm)



import weakref
from operator import attrgetter

from cached_property import cached_property

from ops.framework import (
    Object,
)

from . import schemas


class VersionedInterface(Object):
    def __init__(self, charm, relation_name):
        super().__init__(charm, relation_name)
        self.charm = weakref.proxy(charm)
        self.relation_name = relation_name

        # Future-proof against the need to evolve the relation protocol
        # by ensuring that we agree on a version number before starting.
        # This may or may not be made moot by a future feature in Juju.
        self._set_version()

    def _set_version(self):
        if self.unit.is_leader():
            for relation in self.model.relations.get(self.relation_name, []):
                relation.data[self.app]["version"] = str(schemas.max_version)

    @cached_property
    def relations(self):
        relations = self.model.relations.get(self.relation_name, [])
        return [
            relation
            for relation in sorted(relations, key=attrgetter("id"))
            if self._schema(relation)
        ]

    def _schema(self, relation=None):
        if relation is None:
            return schemas.versions[schemas.max_version]
        if relation.app not in relation.data:
            return None
        data = relation.data[relation.app]
        if "version" not in data:
            return None
        remote_version = int(data["version"])
        return schemas.versions[min((schemas.max_version, remote_version))]

    @property
    def model(self):
        return self.framework.model

    @property
    def app(self):
        return self.charm.app

    @property
    def unit(self):
        return self.charm.unit



#!/usr/bin/env python3

import logging

from ops.charm import CharmBase
from ops.main import main
from ops.model import ActiveStatus, MaintenanceStatus

from loadbalancer_interface import LBConsumers


log = logging.getLogger(__name__)


class ProvidesOperatorCharm(CharmBase):
    def __init__(self, *args):
        super().__init__(*args)
        self.lb_consumers = LBConsumers(self, "lb-consumers")

        self.framework.observe(self.lb_consumers.on.requests_changed, self._provide_lbs)
        self.unit.status = ActiveStatus()

    def _provide_lbs(self, event):
        self.unit.status = MaintenanceStatus("processing requests")
        for request in self.lb_consumers.new_requests:
            response = request.response
            if not request.public:
                response.error = response.error_types.unsupported
                response.error_fields = {"public": "public only"}
            else:
                try:
                    response.address = self._create_lb(request)
                except Exception as e:
                    log.exception("Error creating load balancer")
                    response.error = response.error_types.provider_error
                    response.error_message = str(e)
            self.lb_consumers.send_response(request)
        self.unit.status = ActiveStatus()

    def _create_lb(self, request):
        return "dummy-lb"


if __name__ == "__main__":
    main(ProvidesOperatorCharm)



#!/usr/bin/env python3

from charms.reactive import when, when_not, set_flag, endpoint_from_name
from charms import layer


@when_not("charm.status.is-set")
def set_status():
    layer.status.active("")
    set_flag("charm.status.is-set")


@when("endpoint.lb-consumers.requests_changed")
def get_lb():
    layer.status.maintenance("processing requests")
    lb_consumers = endpoint_from_name("lb-consumers")
    for request in lb_consumers.new_requests:
        response = request.response
        if not request.public:
            response.error = response.error_types.unsupported
            response.error_fields = {"public": "public only"}
        else:
            try:
                response.address = layer.provides_reactive.create_lb(request)
            except Exception as e:
                response.error = response.error_types.provider_error
                response.error_message = str(e)
        lb_consumers.send_response(request)
    layer.status.active("")



#!/usr/bin/env python3

from charms.reactive import set_flag, clear_flag, when, when_not, endpoint_from_name
from charmhelpers.core import hookenv
from charms import layer


@when("endpoint.lb-provider.joined")
@when_not("endpoint.lb-provider.available")
def waiting():
    layer.status.waiting("waiting for provider")


@when("endpoint.lb-provider.available")
@when_not("charm.request.sent")
def request_lb():
    layer.status.maintenance("sending request")
    lb_provider = endpoint_from_name("lb-provider")
    request = lb_provider.get_request("my-service")
    request.protocol = request.protocols.https
    request.port_mapping = {443: 443}
    request.public = hookenv.config("public")
    lb_provider.send_request(request)
    set_flag("charm.request.sent")


@when("charm.request.sent")
@when_not("endpoint.lb-provider.response.available")
def waiting_for_response():
    layer.status.waiting("waiting for provider repsonse")


@when("charm.request.sent")
@when("config.changed.public")
def update_request():
    clear_flag("charm.request.sent")
    clear_flag("config.changed.public")


@when("endpoint.lb-provider.response.changed")
def get_lb():
    lb_provider = endpoint_from_name("lb-provider")
    response = lb_provider.get_response("my-service")
    if response.error:
        layer.status.blocked(f"LB failed: {response.error}")
        hookenv.log(
            f"LB failed ({response.error}):\n"
            f"{response.error_message}\n"
            f"{response.error_fields}",
            hookenv.ERROR,
        )
        return
    hookenv.log(f"LB is available at {response.address}")
    lb_provider.ack_response(response)
    layer.status.active(response.address)



import json
from enum import Enum
from marshmallow import (
    Schema,
    fields,
    validates_schema,
    ValidationError,
)
from marshmallow_enum import EnumField

from .base import SchemaWrapper


version = 1


class Protocols(Enum):
    tcp = "tcp"
    udp = "udp"
    http = "http"
    https = "https"

    def __str__(self):
        return self.value


class ErrorTypes(Enum):
    unsupported = "unsupported"
    provider_error = "provider error"

    def __str__(self):
        return self.value


class Response(SchemaWrapper):
    error_types = ErrorTypes

    class _Schema(Schema):
        error = EnumField(ErrorTypes, missing=None)
        error_message = fields.Str(missing=None)
        error_fields = fields.Dict(key=fields.Str, value=fields.Str, missing=dict)
        address = fields.Str(missing=None)
        received_hash = fields.Str(missing=None)

        @validates_schema
        def _validate(self, data, **kwargs):
            if not data["error"] and not data["address"]:
                raise ValidationError("address required on success")
            if data["error"] and not (data["error_message"] or data["error_fields"]):
                raise ValidationError(
                    "error_message or error_fields required on failure"
                )
            request_fields = Request._Schema().fields
            unknown_fields = data["error_fields"].keys() - request_fields.keys()
            if unknown_fields:
                s = "s" if len(unknown_fields) > 1 else ""
                raise ValidationError(
                    {
                        "error_fields": "Unknown field{}: {}".format(
                            s, ", ".join(unknown_fields)
                        )
                    }
                )

    def __init__(self, request):
        super().__init__()
        self._name = request.name

    @property
    def name(self):
        return self._name

    def __bool__(self):
        return self.hash is not None


class HealthCheck(SchemaWrapper):
    class _Schema(Schema):
        protocol = EnumField(Protocols, by_value=True, required=True)
        port = fields.Int(required=True)
        path = fields.Str(missing=None)
        interval = fields.Int(missing=30)
        retries = fields.Int(missing=3)


class HealthCheckField(fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):
        return value._schema.dump(value)

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, HealthCheck):
            return value
        return HealthCheck()._update(value)


class Request(SchemaWrapper):
    protocols = Protocols

    class _Schema(Schema):
        id = fields.Str(required=True)
        name = fields.Str(required=True)
        protocol = EnumField(Protocols, by_value=True, required=True)
        backends = fields.List(fields.Str(), missing=list)
        port_mapping = fields.Dict(
            keys=fields.Int(), values=fields.Int(), required=True
        )
        algorithm = fields.List(fields.Str(), missing=list)
        sticky = fields.Bool(missing=False)
        health_checks = fields.List(HealthCheckField, missing=list)
        public = fields.Bool(missing=True)
        tls_termination = fields.Bool(missing=False)
        tls_cert = fields.Str(missing=None)
        tls_key = fields.Str(missing=None)
        ingress_address = fields.Str(missing=None)
        sent_hash = fields.Str(missing=None)

    def __init__(self):
        super().__init__()
        self._response = None
        # On the provider side, requests need to track which relation they
        # came from to know where to send the response.
        self.relation = None

    @property
    def response(self):
        if self._response is None:
            self._response = Response(self)
        return self._response

    @classmethod
    def loads(cls, request_sdata, response_sdata=None):
        self = cls()
        self._update(json.loads(request_sdata))
        if response_sdata:
            self.response._update(json.loads(response_sdata))
        return self

    def add_health_check(self, **kwargs):
        """Create a HealthCheck and add it to the list."""
        health_check = HealthCheck()._update(kwargs)
        self.health_checks.append(health_check)
        return health_check



from setuptools import setup
from shutil import rmtree


SETUP = {
    "name": "loadbalancer_interface",
    "version": "1.0.6",
    "author": "Cory Johns",
    "author_email": "cory.johns@canonical.com",
    "url": "https://github.com/juju-solutions/loadbalancer-interface",
    "packages": [
        "loadbalancer_interface",
        "loadbalancer_interface.schemas",
        "loadbalancer_interface.examples",  # Synthetic package populated at build time
    ],
    "package_dir": {
        "loadbalancer_interface.examples": "examples",
    },
    "package_data": {
        "loadbalancer_interface.examples": [
            "../examples/lb-provider/*",
            "../examples/lb-provider/*/*",
            "../examples/lb-consumer/*",
            "../examples/lb-consumer/*/*",
            "../examples/lb-provider-reactive/*",
            "../examples/lb-provider-reactive/*/*",
            "../examples/lb-provider-reactive/*/*/*/*",
            "../examples/lb-consumer-reactive/*",
            "../examples/lb-consumer-reactive/*/*",
            "../examples/lb-consumer-reactive/*/*/*/*",
        ],
    },
    "install_requires": [
        "ops>=1.0.0",
        "cached_property",
        "marshmallow",
        "marshmallow-enum",
        "ops_reactive_interface",
    ],
    "entry_points": {
        "ops_reactive_interface.provides": "loadbalancer = loadbalancer_interface:LBConsumers",  # noqa
        "ops_reactive_interface.requires": "loadbalancer = loadbalancer_interface:LBProvider",  # noqa
        "pytest11": [
            "loadbalancer-test-charm = loadbalancer_interface.pytest_plugin",
        ]
    },
    "license": "Apache License 2.0",
    "description": "'loadbalancer' interface protocol API library",
    "long_description_content_type": "text/markdown",
    "long_description": open("README.md").read(),
}


if __name__ == "__main__":
    rmtree("dist", ignore_errors=True)
    setup(**SETUP)
    rmtree("loadbalancer_interface.egg-info", ignore_errors=True)



import inspect
import json
from hashlib import md5

from marshmallow import (
    missing,
    Schema,
    ValidationError,
)


class SchemaWrapper:
    class _Schema(Schema):
        pass

    def __init__(self):
        self._schema = self._Schema()
        self.version = inspect.getmodule(self).version
        for field_name, field in self._schema.fields.items():
            if field.missing is not missing:
                value = field.missing
                if callable(value):
                    value = value()
            else:
                value = None
            setattr(self, field_name, value)

    def _update(self, data=None, **kwdata):
        if data is None:
            data = {}
        data.update(kwdata)
        for field, value in self._schema.load(data).items():
            setattr(self, field, value)
        return self

    def dump(self):
        # We have to manually validate every field first, or serialization can
        # can fail and we won't know which field it failed on.
        for field_name, field in self._schema.fields.items():
            value = getattr(self, field_name, None)
            try:
                if hasattr(field, "_validated"):
                    # For some reason, some field types do their validation in
                    # a `_validated` method, rather than in the `_validate`
                    # from the base Field class. For those, calling `_validate`
                    # doesn't actually do any validation.
                    field._validated(value)
                else:
                    field._validate(value)
                field._validate_missing(value)
            except ValidationError as e:
                raise ValidationError({field_name: e.messages}) from e
        serialized = self._schema.dump(self)
        # Then we have to validate the serialized data again to catch any
        # schema-level validation issues.
        errors = self._schema.validate(serialized)
        if errors:
            raise ValidationError(errors)
        return serialized

    def dumps(self):
        return json.dumps(self.dump(), sort_keys=True)

    @property
    def hash(self):
        try:
            return md5(self.dumps().encode("utf8")).hexdigest()
        except ValidationError:
            return None
